import sys
import pandas
import argparse

from shipyard_bp_utils import files
from shipyard_bp_utils.artifacts import Artifact
from shipyard_airtable import AirtableClient
from shipyard_bp_utils.args import convert_to_boolean
from shipyard_templates import ExitCodeException, ShipyardLogger

logger = ShipyardLogger.get_logger()


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--api-key", dest="api_key", required=True)
    parser.add_argument("--base-id", dest="base_id", required=True)
    parser.add_argument("--table-id", dest="table_id", required=True)
    parser.add_argument(
        "--filename-or-pattern", dest="filename_or_pattern", required=True
    )
    parser.add_argument(
        "--source-folder-name", dest="source_folder_name", required=False, default=""
    )
    parser.add_argument(
        "--source-file-name-match-type",
        dest="source_file_name_match_type",
        required=True,
        default="exact_match",
    )
    parser.add_argument("--key-fields", dest="key_fields", required=False)
    parser.add_argument("--typecast", dest="typecast", required=False, default="FALSE")
    parser.add_argument("--insert-method", dest="insert_method", required=True)
    return parser.parse_args()


def prepare_data_from_csv(file):
    logger.debug(f"Preparing data from file: {file}")
    records = []
    df = pandas.read_csv(file)

    for index, row in df.iterrows():
        record = {"fields": row.to_dict()}
        if "id" in record["fields"]:
            del record["fields"]["id"]
        records.append(record)
    logger.debug(f"Following records are prepared: \n {records}")
    return records


def main():
    try:
        artifact = Artifact("airtable")
        args = get_args()

        typecast = convert_to_boolean(args.typecast) if args.typecast else False
        search_for = args.filename_or_pattern
        match_type = args.source_file_name_match_type
        source_folder = args.source_folder_name or "."
        key_fields = args.key_fields = args.key_fields.split(",")
        upload_method = args.insert_method.lower()

        files_found = files.find_matching_files(
            args.filename_or_pattern, source_folder, match_type
        )

        client = AirtableClient(api_key=args.api_key)
        if upload_method == "replace":
            client.clear_table(args.base_id, args.table_id)
            upload_method = "append"

        if len(files_found) == 0:
            logger.error(
                f"No files found matching search criteria: {search_for} in {source_folder} "
                f"with match type: {match_type}"
            )
            sys.exit(client.EXIT_CODE_FILE_NOT_FOUND)

        failed, success = [], []
        logger.info(f"Uploading {len(files_found)} file(s) to Airtable...")

        for file in files_found:

            try:
                upload_args = {
                    "base": args.base_id,
                    "table": args.table_id,
                    "data": prepare_data_from_csv(file),
                    "typecast": typecast,
                    "upload_method": upload_method,
                }
                if key_fields:
                    upload_args["key_fields"] = key_fields
                client.upload(**upload_args)

            except Exception as e:
                logger.error(f"Error uploading file: {file} to Airtable: {e}")
                failed.append(
                    {
                        "file": file,
                        "error": f"Error uploading file: {file} to Airtable: {e}",
                    }
                )

            else:
                success.append(file)
                logger.info(f"Uploaded {file} to Airtable successfully.")

    except ExitCodeException as err:
        logger.error(err.message)
        sys.exit(err.exit_code)
    except Exception as e:
        logger.error(f"Unexpected error occurred: {e}")
        sys.exit(1)

    else:
        artifact.logs.write_json(
            "airtable_upload_status", {"success": success, "failed": failed}
        )
        if success:
            logger.info(f"Uploaded {len(success)} file(s) to Airtable successfully.")
        if failed:
            logger.error(f"Failed to upload {len(failed)} file(s) to Airtable.")
            sys.exit(client.EXIT_CODE_UPLOAD_ERROR)


if __name__ == "__main__":
    main()



from pyairtable import Api
from requests.exceptions import HTTPError
from shipyard_templates import Spreadsheets, ShipyardLogger, ExitCodeException

logger = ShipyardLogger.get_logger()


class AirtableClient(Spreadsheets):
    EXIT_CODE_INVALID_CREDENTIALS = 100
    EXIT_CODE_INVALID_BASE = 101
    EXIT_CODE_INVALID_TABLE = 102
    EXIT_CODE_INVALID_VIEW = 103
    EXIT_CODE_RESOURCE_NOT_FOUND = 104

    def __init__(self, api_key: str) -> None:
        self.api = Api(api_key)

    def _handle_error(self, error: HTTPError) -> None:
        """Handle HTTPError and raise ExitCodeException with appropriate message and exit code.

        Args:
            error (HTTPError): The HTTPError object.

        Raises:
            ExitCodeException: The exception with the appropriate message and exit code.
        """
        logger.debug("Handling HTTPError...")
        logger.debug(f"Error: {error.response.status_code} {error.response.reason}")
        response_details = error.response.json()
        error_details = response_details.get("error", {})

        if error.response.status_code == 401:
            raise ExitCodeException(
                error_details.get(
                    "message",
                    "The API key provided is incorrect."
                    "Please double-check for extra spaces or invalid characters.",
                ),
                self.EXIT_CODE_INVALID_CREDENTIALS,
            )
        elif error.response.status_code == 403:
            raise ExitCodeException(
                error_details.get(
                    "message",
                    "The provided key does not have access to the requested resource."
                    "Please check the scope permissions of the API key.",
                ),
                self.EXIT_CODE_INVALID_CREDENTIALS,
            )
        elif error_details.type == "TABLE_NOT_FOUND":
            raise ExitCodeException(
                error_details.get("message", "The requested table was not found."),
                self.EXIT_CODE_INVALID_TABLE,
            )
        elif error_details.type == "VIEW_NAME_NOT_FOUND":
            raise ExitCodeException(
                error_details.get("message", "The requested view was not found."),
                self.EXIT_CODE_INVALID_VIEW,
            )

        elif error.response.status_code == 404:
            raise ExitCodeException(
                error_details.get("message", "The requested resource was not found."),
                self.EXIT_CODE_RESOURCE_NOT_FOUND,
            )
        elif error.response.status_code == 429:
            raise ExitCodeException(
                error_details.get(
                    "message", "The rate limit for the API key has been exceeded."
                ),
                self.EXIT_CODE_RATE_LIMIT,
            )
        else:
            raise ExitCodeException(
                error_details.get("message", "An unexpected error occurred."),
                self.EXIT_CODE_UNKNOWN_ERROR,
            )

    def connect(self) -> int:
        """Test the connection to Airtable.

        Returns:
            int: The exit code.
        """

        try:
            self.api.whoami()
        except HTTPError as err:
            if err.response.status_code == 401:
                logger.authtest("Invalid Personal Access Token")
            elif err.response.status_code == 403:
                logger.authtest(
                    "The provided Personal Access Token does not have access to the requested resource."
                    "Please check the permissions."
                )

            else:
                logger.authtest(
                    f"Failed: {err.response.status_code} {err.response.reason}"
                )
            return 1
        else:
            logger.authtest("Success")
            return 0

    def fetch(self, base: str, table: str, view: str = None) -> list:
        """
        Fetch data from Airtable.

        Args:
            base (str): The base ID.
            table (str): The table name.
            view (str, optional): The view name. Defaults to None.

        Returns:
            list: The list of records.
        """
        logger.debug("Fetching data from Airtable...")
        try:
            table = self.api.table(base, table)
            records = table.all(view=view) if view else table.all()
        except HTTPError as err:
            self._handle_error(err)
        else:
            logger.debug(f"Returned {len(records)} record(s)")
            return records

    def upload(
        self,
        upload_method: str,
        base: str,
        table: str,
        data: list,
        key_fields: list = None,
        typecast: bool = True,
    ) -> None:
        """
        Upload data to Airtable.

        Args:
            upload_method (str): The method to use for uploading the data. Choose from 'append', 'upsert', or 'replace'.
            base (str): The base ID.
            table (str): The table name.
            data (list): The list of records to upload.
            key_fields (list, optional): The list of fields to use as keys. Defaults to None.
            typecast (bool, optional): Whether to typecast the data. Defaults to True.
        """

        upload_method = upload_method.lower()
        if upload_method == "append":
            self.batch_create_records(base, table, data, key_fields, typecast)
        elif upload_method == "upsert":
            self.batch_upsert_records(base, table, data, key_fields, typecast)
        else:
            raise ExitCodeException(
                "Invalid upload method. Please choose from 'append', 'upsert', or 'replace'",
                self.EXIT_CODE_INVALID_INPUT,
            )

    def batch_create_records(
        self, base: str, table: str, data: list, key_fields: list, typecast: bool = True
    ) -> None:
        """
        Create records in Airtable.

        Args:
            base (str): The base ID.
            table (str): The table name.
            data (list): The list of records to create.
            key_fields (list): The list of fields to use as keys.
            typecast (bool, optional): Whether to typecast the data. Defaults to True.
        """
        logger.debug("Inserting data to Airtable...")
        try:
            table = self.api.table(base, table)
            table.batch_upsert(records=data, key_fields=key_fields, typecast=typecast)
        except HTTPError as err:
            self._handle_error(err)
        else:
            logger.debug("Data inserted successfully")

    def batch_upsert_records(
        self, base: str, table: str, data: list, key_fields: list, typecast: bool = True
    ) -> None:
        """
        Upsert data to Airtable.

        Args:
            base (str): The base ID.
            table (str): The table name.
            data (list): The list of records to upsert.
            key_fields (list): The list of fields to use as keys.
            typecast (bool, optional): Whether to typecast the data. Defaults to True.

        """

        logger.debug("Upserting data to Airtable...")
        try:
            table = self.api.table(base, table)
            table.batch_upsert(records=data, key_fields=key_fields, typecast=typecast)
        except HTTPError as err:
            self._handle_error(err)
        else:
            logger.debug("Data uploaded successfully")

    def clear_table(self, base: str, table: str) -> None:
        """
        Delete all records from a table.

        Args:
            base (str): The base ID.
            table (str): The table name.
        """
        logger.debug("Clearing table...")
        try:
            table = self.api.table(base, table)
            records = table.all()
            record_ids = [record["id"] for record in records]
            table.batch_delete(record_ids)

        except HTTPError as err:
            self._handle_error(err)
        else:
            logger.debug("Table cleared successfully")
